<?php
namespace Javorszky\Toolbox\Utilities\Process;

use Javorszky\Toolbox\Utilities\Validate as Validate;
use Javorszky\Toolbox\Utilities as Utilities;
use WCS_Download_Handler;
use DateTime;
use DateTimeZone;

/**
 * Recalculates the next date after the current next date. We need to adjust a filter, so calculate_date will base
 * the calculation on the currently set "next_payment" date.
 *
 * @param  integer                  $user_id    ID of the current user
 * @param  \WC_Subscription          $subscription Subscription we're shipping and keeping the date for
 * @param  string                   $nonce      a security nonce
 * @return boolean                              whether anything happened
 */
function process_skip_next_date( $user_id, $subscription, $nonce ) {
	if ( ! wcs_is_subscription( $subscription ) ) {
		// there's nothing to do
		return false;
	}

	$old_next_timestamp = $subscription->get_time( 'next_payment' );
	$old_next_date      = $subscription->get_date( 'next_payment' );

	if (
		Validate\validate_subscription_ownership( $user_id, $subscription )
		&& Validate\validate_skip_next_request( $subscription, $old_next_timestamp, $nonce )
		&& apply_filters( 'jgtb_validate_skip_next_request', true, $subscription, $user_id, $old_next_timestamp )
		&& 'yes' === apply_filters( 'jgtb_allow_edit_date_for_subscription', 'yes', $subscription )
		) {
		add_filter( 'wcs_calculate_next_payment_from_last_payment', '__return_false' );
		$next_payment_date = $subscription->calculate_date( 'next_payment' );
		remove_filter( 'wcs_calculate_next_payment_from_last_payment', '__return_false' );

		$next_payment_date = apply_filters( 'sptb_next_payment_date_after_skip', $next_payment_date, $subscription );

		$subscription->update_dates( array( 'next_payment' => $next_payment_date ) );
		$subscription->add_order_note( sprintf( "Customer chose to skip the next shipping and has set the next payment date \nfrom %s\nto %s.", $old_next_date, $next_payment_date ), false, true );

		// save skip meta to hide button in future
		$subscription->update_meta_data( 'wcs_toolbox_skipped', 1 );

		$subscription->save();

		do_action( 'jgtb_after_skip_next_date', $subscription, $user_id, $old_next_date, $next_payment_date );

		return true;
	}
	return false;
}

/**
 * A central place to handle processing the ship now keep date request for a subscription. Protected, so it can be
 * accessed from children classes too, but not outside, and handles all validations. Assuming nonce generated
 * with the base of $subscription->id . '_completed_' . $completed_payments.
 *
 * @param  integer                  $user_id    ID of the current user
 * @param  \WC_Subscription          $subscription Subscription we're shipping and keeping the date for
 * @param  string                   $nonce      a security nonce
 * @return boolean                              true if something happened
 */
function process_ship_now_keep_date( $user_id, $subscription, $nonce ) {
	if ( ! wcs_is_subscription( $subscription ) ) {
		// there's nothing to do
		return false;
	}

	$completed_payments = $subscription->get_payment_count( 'completed' );

	if (
		Validate\validate_subscription_ownership( $user_id, $subscription )
		&& Validate\validate_ship_now_keep_date( $subscription, $completed_payments, $nonce )
		&& apply_filters( 'jgtb_validate_ship_now_keep_date', true, $subscription, $user_id, $completed_payments )
		&& 'yes' === apply_filters( 'jgtb_allow_edit_date_for_subscription', 'yes', $subscription )
		) {
		Utilities\ship_now( $subscription );
		$subscription->add_order_note( sprintf( 'Customer created a new renewal order and chose to keep the next payment date as is: %s.', $subscription->get_date( 'next_payment' ) ), false, true );

		do_action( 'jgtb_after_ship_now_keep_date', $subscription, $user_id );

		return true;
	}
	return false;

}

/**
 * A central place to handle processing the ship now adjust date request for a subscription. Protected, so it can be
 * accessed from children classes too, but not outside, and handles all validations. Assuming nonce generated
 * with the base of $subscription->id . '_completed_' . $completed_payments.
 *
 * @param  integer                  $user_id    ID of the current user
 * @param  \WC_Subscription          $subscription Subscription we're shipping and keeping the date for
 * @param  string                   $nonce      a security nonce
 * @return boolean                              whether anything successful happened
 */
function process_ship_now_adjust_date( $user_id, $subscription, $nonce ) {
	if ( ! wcs_is_subscription( $subscription ) ) {
		// there's nothing to do
		return false;
	}

	$old_next_date      = $subscription->get_date( 'next_payment' );
	$completed_payments = $subscription->get_payment_count( 'completed' );

	if (
		Validate\validate_subscription_ownership( $user_id, $subscription )
		&& Validate\validate_ship_now_adjust_date( $subscription, $completed_payments, $nonce )
		&& apply_filters( 'jgtb_validate_ship_now_adjust_date', true, $subscription, $user_id, $completed_payments )
		&& 'yes' === apply_filters( 'jgtb_allow_edit_date_for_subscription', 'yes', $subscription )
		) {
		Utilities\ship_now( $subscription );
		$next_payment_date = Utilities\adjust_date( $subscription, $old_next_date );
		$subscription->add_order_note( sprintf( "Customer created a renewal order, and chose to have the next payment date rescheduled with relation to today\nfrom %s\nto %s.", $old_next_date, $next_payment_date ), false, true );

		do_action( 'jgtb_after_ship_now_adjust_date', $subscription, $user_id, $old_next_date, $next_payment_date );

		return true;
	}
	return false;
}

/**
 * One central place to change the shipping address of a specific subscription. Nonce is in
 * $_REQUEST['jgtb_edit_details_of_' . $subscription->id] and is generated by
 * wcs_edit_details_of_ . $subscription->id. Address details are sieved from $_REQUEST by
 * $this->sieve_address_details and then a difference is calculated to only update what changed.
 *
 * @param  integer                  $user_id    ID of the current user
 * @param  \WC_Subscription          $subscription Subscription we're shipping and keeping the date for
 * @param  string                   $nonce      a security nonce
 * @return boolean                              whether anything happened
 */
function process_change_shipping_address( $user_id, $subscription, $nonce ) {
	if (
		Validate\validate_subscription_ownership( $user_id, $subscription )
		&& Validate\validate_edit_details( $subscription, $nonce )
		&& apply_filters( 'jgtb_validate_edit_details', true, $subscription, $user_id )
		&& apply_filters( 'jgtb_validate_change_shipping_address', true, $subscription, $user_id )
		) {
		$old_shipping_addresses = $subscription->get_address( 'shipping' );

		$new_shipping_address = Utilities\sieve_address_details( $_REQUEST, 'shipping' ); // phpcs:ignore CSRF okay.

		$shipping_diff = Utilities\get_difference( $old_shipping_addresses, $new_shipping_address );

		do_action( 'jgtb_change_shipping_address', $old_shipping_addresses, $new_shipping_address, $shipping_diff, $user_id, $subscription );

		if ( ! empty( $shipping_diff ) ) {
			$shipping_address_string = '';
			foreach ( $shipping_diff as $key => $value ) {
				$shipping_address_string .= PHP_EOL . '[' . $key . ']: ' . $value;
			}
			$subscription->set_address( $shipping_diff, 'shipping' );
			$subscription->add_order_note( sprintf( 'Customer chose to update the shipping details with: %s', $shipping_address_string ), false, true );

			// Update subscription last modified date
			Utilities\update_post_modified_date( $subscription );

			return true;
		}
	}
	return false;
}

/**
 * One central place to change the billing address of a specific subscription. Nonce is in
 * $_REQUEST['jgtb_edit_details_of_' . $subscription->id] and is generated by
 * wcs_edit_details_of_ . $subscription->id. Address details are sieved from $_REQUEST by
 * $this->sieve_address_details and then a difference is calculated to only update what changed.
 *
 * @param  integer                  $user_id    ID of the current user
 * @param  \WC_Subscription          $subscription Subscription we're shipping and keeping the date for
 * @param  string                   $nonce      a security nonce
 * @return boolean                              whether anything happened
 */
function process_change_billing_address( $user_id, $subscription, $nonce ) {
	if (
		Validate\validate_subscription_ownership( $user_id, $subscription )
		&& Validate\validate_edit_details( $subscription, $nonce )
		&& apply_filters( 'jgtb_validate_edit_details', true, $subscription, $user_id )
		&& apply_filters( 'jgtb_validate_change_billing_address', true, $subscription, $user_id )
		) {
		$old_billing_addresses = $subscription->get_address( 'billing' );

		$new_billing_address = Utilities\sieve_address_details( $_REQUEST, 'billing' ); // phpcs:ignore CSRF ok

		$billing_diff = Utilities\get_difference( $old_billing_addresses, $new_billing_address );

		do_action( 'jgtb_change_billing_address', $old_billing_addresses, $new_billing_address, $billing_diff, $user_id, $subscription );

		if ( ! empty( $billing_diff ) ) {
			$billing_address_string = '';
			foreach ( $billing_diff as $key => $value ) {
				$billing_address_string .= PHP_EOL . '[' . $key . ']: ' . $value;
			}

			$subscription->set_address( $billing_diff, 'billing' );
			$subscription->add_order_note( sprintf( 'Customer chose to update the billing details with: %s', $billing_address_string ), false, true );

			// Update subscription last modified date
			Utilities\update_post_modified_date( $subscription );

			return true;
		}
	}
	return false;
}

/**
 * One central place to change the product quantities of a specific subscription. Nonce is in
 * $_REQUEST['jgtb_edit_details_of_' . $subscription->id], and is generated by
 * wcs_edit_details_of_ . $subscription->id.
 *
 * @param integer                   $user_id
 * @param \WC_Subscription          $subscription
 * @param string                    $nonce
 * @return boolean                  whether anything successful happened
 * @throws \Exception
 */
function process_change_product_quantities( $user_id, $subscription, $nonce ) {

	if (
		Validate\validate_subscription_ownership( $user_id, $subscription )
		&& Validate\validate_edit_details( $subscription, $nonce )
		&& apply_filters( 'jgtb_validate_edit_details', true, $subscription, $user_id )
		&& apply_filters( 'jgtb_validate_change_product_quantities', true, $subscription, $user_id )
		&& 'yes' === apply_filters( 'jgtb_allow_edit_qty_for_subscription', 'yes', $subscription )
		) {
		// edit the quantity changes
		$qty_changed        = false;
		$subscription_items = $subscription->get_items();
		$changed            = array();
		if ( count( $subscription_items ) > 0 ) {
			foreach ( $subscription_items as $item_id => $item ) {
				$new_qty = ( isset( $_REQUEST[ 'new_quantity_' . $item_id ] ) ) ? absint( $_REQUEST[ 'new_quantity_' . $item_id ] ) : null; // phpcs:ignore CSRF ok
				$old_qty = $item->get_quantity();
				
				if ( null !== $new_qty && $new_qty !== $old_qty ) { // phpcs:ignore Standard.Category.SniffName.ErrorCode
					$qty_changed         = true;
					$product_id          = wcs_get_canonical_product_id( $item );
					$product             = wc_get_product( absint( $product_id ) );
					$item_name           = wcs_get_line_item_name( $item );
					$changed[ $item_id ] = array(
						'product_id' => $product_id,
						'old'        => $old_qty,
						'new'        => $new_qty,
					);

					if ( 0 === absint( $new_qty ) ) {
						// remove the item
						WCS_Download_Handler::revoke_downloadable_file_permission( $product_id, $subscription->get_id(), $subscription->get_user_id() );

						wc_update_order_item( $item_id, array( 'order_item_type' => 'line_item_removed' ) );

						$subscription->add_order_note( sprintf( 'Customer removed "%1$s" (Product ID: #%2$d) via the Edit Subscription Details page.', $item_name, $product_id ) );
					}

					// set prices
					$item->set_subtotal( ( $item->get_subtotal() / $old_qty ) * $new_qty );
					$item->set_total( ( $item->get_total() / $old_qty ) * $new_qty );

					// set quantity
					$item->set_quantity( $new_qty );

					// save
					$item->save();

					$subscription->add_order_note( sprintf( 'Customer changed quantity of "%1$s" from %2$s to %3$s.', $item_name, $old_qty, $new_qty ) );
				}
			}
		}

		if ( $qty_changed ) {
			// Maybe recalculate shipping for the subscription.
			Utilities\maybe_recalculate_and_apply_shipping_for_subscription( $subscription );

			// recalculate things
			$subscription->calculate_totals();
			$subscription->save();

			// need to differentiate between 3 things: this is succeeded
			do_action( 'jgtb_after_change_product_quantities', $subscription, $changed );

			// Update subscription last modified date
			Utilities\update_post_modified_date( $subscription );

			return 1;
		}

		do_action( 'jgtb_after_not_changed_product_quantities', $subscription, $changed );

		$subscription->save();

		// need to differentiate between 3 things: this is succeeded in not changing anything
		return 0;
	}

	// need to differentiate between 3 things: this is failed
	return -1;
}

/**
 * Change frequency. Hooked into priority 30, because this needs to happen after the change frequency processing is
 * done. Otherwise this would first change the next payment date of the subscription making the nonce that relied on
 * the previous next payment date invalid for the change next payment date function which would throw a notice then.
 *
 * I've gotten around that two ways:
 * 1. first change the next payment date, as the nonce for this function does not depend on the next payment date, so
 * both will successfully run, and
 * 2. only run the change next payment date function if the user changed the next payment date.
 *
 * If the user changed the frequency, the next payment date will be recalculated from today anyways though, so in that
 * sense any change that they've done to the next payment date will be irrelevant and they'd have to do it again.
 *
 * @param integer          $user_id      ID of the current user
 * @param \WC_Subscription $subscription Subscription we're changing the frequency for
 * @param string           $nonce        a security nonce
 * @param integer          $new_interval a number between 1-6 inclusive by default
 * @param string           $new_period   day / week / month / year by default
 *
 */
function process_change_frequency( $user_id, $subscription, $nonce, $new_interval, $new_period ) {
	if (
		Validate\validate_subscription_ownership( $user_id, $subscription )
		&& Validate\validate_change_frequency_request( $subscription, $nonce )
		&& apply_filters( 'jgtb_validate_change_frequency', true, $subscription, $user_id, $new_interval, $new_period )
		&& 'yes' === apply_filters( 'jgtb_allow_edit_date_for_subscription', 'yes', $subscription )
		) {
		$next_ship_date = $subscription->get_date( 'next_payment' );

		$old_interval = $subscription->get_billing_interval();
		$old_period   = $subscription->get_billing_period();

		if ( $new_interval != $old_interval ) { // phpcs:ignore loose comparison ok.
			$subscription->set_billing_interval( $new_interval );
		}

		if ( $new_period != $old_period ) { // phpcs:ignore loose comparison ok.
			$subscription->set_billing_period( $new_period );
		}

		/**
		 * Filter whether to change the next_payment date of the subscription when frequency is changed.
		 *
		 * @since 1.4.22
		 *
		 * @param bool $update_next_date default true
		 * @param int $user_id ID of the user making the change
		 * @param \WC_Subscription $subscription Subscription the frequency is changing
		 */
		if ( true === (bool) apply_filters( 'sptb_frequency_change_next_payment', true, $user_id, $subscription ) ) {
			$next_ship_date = gmdate( 'Y-m-d H:i:s', strtotime( '+' . $new_interval . ' ' . $new_period ) );
			$subscription->update_dates( array( 'next_payment' => $next_ship_date ) );
		}

		$subscription->add_order_note( sprintf( 'Customer chose to change the frequency' . PHP_EOL . 'from %1$s %2$s' . PHP_EOL . 'to %3$s %4$s.', $old_interval, $old_period, $new_interval, $new_period ), false, true );
		$subscription->save();

		do_action( 'jgtb_after_change_frequency', $subscription, $old_interval, $old_period, $new_interval, $new_period, $next_ship_date );

		wc_add_notice( apply_filters( 'jgtb_frequency_updated_message', __( 'The subscription frequency has been updated!', 'jg-toolbox' ), $subscription ) );

		// Update subscription last modified date
		Utilities\update_post_modified_date( $subscription );
	}
}

/**
 * One central place to change the shipping method of a specific subscription. Nonce is in
 * $_REQUEST['jgtb_change_shipping_method_nonce'] and is generated by
 * 'change_shipping_method_' . $subscription->get_id() . $current_shipping_method.
 *
 * @param  int              $user_id                     ID of the current user
 * @param  \WC_Subscription $subscription                Subscription we're shipping and keeping the date for
 * @param  string           $nonce                       a security nonce
 * @param  string           $new_shipping_method_rate_id the new shipping method rate ID
 * @param  string           $old_shipping_method_rate_id the old shipping method rate ID
 * @return void
 */
function process_change_shipping_method( int $user_id, \WC_Subscription $subscription, string $nonce, string $new_shipping_method_rate_id, string $old_shipping_method_rate_id ): void {
	if (
		Validate\validate_subscription_ownership( $user_id, $subscription )                                                  &&
		Validate\validate_change_shipping_method_request( $subscription, $nonce, $old_shipping_method_rate_id )              &&
		apply_filters( 'jgtb_validate_change_shipping_method', true, $subscription, $user_id, $new_shipping_method_rate_id ) &&
		'yes' === apply_filters( 'jgtb_allow_edit_shipping_method_for_subscription', 'yes', $subscription )
	) {
		// Remove the old method
		foreach ( $subscription->get_items( 'shipping' ) as $item_id => $item ) {
			$subscription->remove_item( $item_id );
		}

		$shipping_rate_found        = false;
		$available_shipping_methods = Utilities\get_subscription_available_shipping_methods( $subscription );
		
		if ( isset( $available_shipping_methods[ $new_shipping_method_rate_id ] ) ) {
			$shipping_method = $available_shipping_methods[ $new_shipping_method_rate_id ];
			
			if ( $shipping_method instanceof \WC_Shipping_Method ) {
				Utilities\add_new_shipping_item_to_subscription( $subscription, $shipping_method );
				$shipping_rate_found = true;
			}
		}

		if ( $shipping_rate_found ) {
			$subscription->calculate_totals();
			$subscription->save();

			wc_add_notice( __( 'Shipping method updated successfully.', 'jg-toolbox' ), 'success' );
		} else {
			wc_add_notice( __( 'Shipping rate not found.', 'jg-toolbox' ), 'error' );
		}

		do_action( 'jgtb_after_change_shipping_method', $subscription, $user_id, $old_shipping_method_rate_id, $new_shipping_method_rate_id );

		// Update subscription last modified date
		Utilities\update_post_modified_date( $subscription );
	}
}

/**
 * One central place to change the shipping method of a specific subscription. Nonce is in
 * $_REQUEST['jgtb_change_shipping_method_nonce'] and is generated by
 * 'change_shipping_method_' . $subscription->get_id() . $current_shipping_method.
 *
 * @param  int              $user_id                     ID of the current user
 * @param  \WC_Subscription $subscription                Subscription we're shipping and keeping the date for
 * @param  string           $nonce                       a security nonce
 * @param  string           $new_shipping_method_rate_id the new shipping method rate ID
 * @param  string           $old_shipping_method_rate_id the old shipping method rate ID
 * @return void
 */
function process_change_shipping_method_legacy( int $user_id, \WC_Subscription $subscription, string $nonce, string $new_shipping_method_rate_id, string $old_shipping_method_rate_id ): void {
	if (
		Validate\validate_subscription_ownership( $user_id, $subscription )                                                  &&
		Validate\validate_change_shipping_method_request( $subscription, $nonce, $old_shipping_method_rate_id )              &&
		apply_filters( 'jgtb_validate_change_shipping_method', true, $subscription, $user_id, $new_shipping_method_rate_id ) &&
		'yes' === apply_filters( 'jgtb_allow_edit_shipping_method_for_subscription', 'yes', $subscription )
	) {
		// Remove the old method
		foreach ( $subscription->get_items( 'shipping' ) as $item_id => $item ) {
			$subscription->remove_item( $item_id );
		}

		$shipping_rate_found = false;

		// Split the rate ID to get the method ID and instance ID
		list( $method_id, $instance_id ) = explode( ':', $new_shipping_method_rate_id );

		// Load default shipping methods
		$default_shipping_methods = \WC()->shipping()->load_shipping_methods();

		// Handle the case for Checkout Block pickup locations
		if ( 'pickup_location' === $method_id ) {
			if ( isset( $default_shipping_methods['pickup_location'] ) && wc_string_to_bool( $default_shipping_methods['pickup_location']->enabled ) ) {
				$pickup_location_pickup_locations = get_option( 'pickup_location_pickup_locations', array() );
				
				if ( isset( $pickup_location_pickup_locations[ $instance_id ] ) && wc_string_to_bool( $pickup_location_pickup_locations[ $instance_id ]['enabled'] ) ) {
					$location      = $pickup_location_pickup_locations[ $instance_id ];
					$shipping_item = new \WC_Order_Item_Shipping();

					$shipping_item->set_props( array(
						'method_title' => sprintf( '%s (%s)', $default_shipping_methods['pickup_location']->method_title, $location['name'] ),
						'method_id'    => $default_shipping_methods['pickup_location']->id,
						'total'        => 0,
						'taxes'        => array(),
						'order_id'     => $subscription->get_id(),
					) );

					$shipping_item->save();
					$subscription->add_item( $shipping_item );
					$shipping_rate_found = true;
				}
			}
		} else {
			// Handle normal shipping methods
			foreach ( Utilities\get_shipping_zones() as $zone_data ) {
				// Check if we're dealing with the default zone or a normal zone
				if ( is_array( $zone_data ) && isset( $zone_data['id'] ) ) {
					$zone = new \WC_Shipping_Zone( $zone_data['id'] );
				} else {
					$zone = $zone_data;
				}

				$zone_shipping_methods = $zone->get_shipping_methods();

				foreach ( $zone_shipping_methods as $shipping_method ) {
					if ( ! isset( $shipping_method->id ) || ! isset( $shipping_method->instance_id ) ) {
						continue;
					}

					if ( $shipping_method->id === $method_id && $shipping_method->instance_id === absint( $instance_id ) ) {
						$shipping_item = new \WC_Order_Item_Shipping();

						$shipping_item->set_props( array(
							'method_title' => $shipping_method->get_title(),
							'method_id'    => $shipping_method->id,
							'total'        => 0,
							'taxes'        => array(),
							'order_id'     => $subscription->get_id(),
						) );

						if ( method_exists( $shipping_method, 'get_rates' ) ) {
							$shipping_rates = $shipping_method->get_rates();

							if ( ! empty( $shipping_rates ) ) {
								$default_rate = reset( $shipping_rates );

								$shipping_item->set_props( array(
									'total' => wc_format_decimal( $default_rate->get_cost() ),
									'taxes' => $default_rate->get_taxes(),
								) );
							}
						}

						$shipping_item->save();
						$subscription->add_item( $shipping_item );
						$shipping_rate_found = true;
						break 2;
					}
				}
			}
		}

		if ( $shipping_rate_found ) {
			$subscription->calculate_totals();
			$subscription->save();

			wc_add_notice( __( 'Shipping method updated successfully.', 'jg-toolbox' ), 'success' );
		} else {
			wc_add_notice( __( 'Shipping rate not found.', 'jg-toolbox' ), 'error' );
		}

		do_action( 'jgtb_after_change_shipping_method', $subscription, $user_id, $old_shipping_method_rate_id, $new_shipping_method_rate_id );

		// Update subscription last modified date
		Utilities\update_post_modified_date( $subscription );
	}
}

/**
 * Function to actually change the shipping date if it changed. Called from RescheduleNextRenewal, only if the user actually changed
 * the date in the input.
 *
 * @param  integer                  $user_id           ID of the current user
 * @param  \WC_Subscription         $subscription      Subscription we're shipping and keeping the date for
 * @param  string                   $nonce             A security nonce
 * @param  string                   $new_ship_date     The new date in Y-m-d format (2018-01-28)
 *
 * @return int                      0 on failure, 1 on success.
 */
function process_change_next_shipping_date( int $user_id, \WC_Subscription $subscription, string $nonce, string $new_ship_date ): int {
	// Validate the request and inputs.
	if (
		! Validate\validate_subscription_ownership( $user_id, $subscription )
		|| ! Validate\validate_change_next_ship_date_request( $subscription, $nonce )
		|| ! Utilities\sanitize_new_date( $new_ship_date )
		|| ! apply_filters( 'jgtb_validate_change_next_ship_date', $subscription, $user_id, $new_ship_date )
		|| ! wc_string_to_bool( apply_filters( 'jgtb_allow_edit_date_for_subscription', 'yes', $subscription ) )
	) {
		return 0;
	}

	if ( apply_filters( 'jgtb_allow_edit_date_for_subscription_needs_payment', $subscription->needs_payment(), $subscription ) ) {
		wc_add_notice( apply_filters( 'jgtb_date_renewal_error_message', __( 'Next shipment date couldn\'t be updated. Subscription needs payment.', 'jg-toolbox' ), $subscription ), 'error' );

		return 0;
	}

	$time_zone     = new DateTimeZone( wc_timezone_string() );
	$old_ship_date = DateTime::createFromFormat( 'Y-m-d H:i:s', $subscription->get_date( 'next_payment', 'site' ), $time_zone );
	$old_end_date  = DateTime::createFromFormat( 'Y-m-d H:i:s', $subscription->get_date( 'end', 'site' ), $time_zone ) ?: null;
	$new_ship_date = DateTime::createFromFormat( 'Y-m-d', esc_attr( $new_ship_date ), $time_zone ); // phpcs:ignore CSRF ok

	// time
	$time = apply_filters( 'jgtb_process_change_next_shipping_date_time', array(
		'hour'   => $old_ship_date->format( 'H' ),
		'minute' => $old_ship_date->format( 'i' ),
		'second' => $old_ship_date->format( 's' ),
	), $old_ship_date, $time_zone, $subscription );

	// Bail only if we don't have valid dates.
	if ( ! $new_ship_date || ! $old_ship_date ) {
		return 0;
	}

	// Validate new ship date.
	if ( is_null( Validate\validate_new_next_shipping_date( $new_ship_date, $old_ship_date, $old_end_date ) ) ) {
		return 0;
	}

	// Set the new shipping time.
	$new_ship_date->setTime( $time['hour'], $time['minute'], $time['second'] );
	$new_ship_date = apply_filters( 'jgtb_new_ship_date', $new_ship_date, $subscription, $user_id );

	$dates = array(
		'next_payment' => $new_ship_date->format( 'Y-m-d H:i:s' ),
	);

	$note = sprintf(
		'<p>Customer chose to change the next shipment date<br>from %s<br>to %s.',
		$old_ship_date->format( 'Y-m-d H:i:s' ),
		$new_ship_date->format( 'Y-m-d H:i:s' )
	);

	$next_payment_after_end_date = get_option( JGTB_OPTION_PREFIX . 'change_next_payment_available_after_end_date', 'no' );
	$adjust_end_date             = apply_filters(
		'jgtb_adjust_end_date_with_next_date',
		wc_string_to_bool( $next_payment_after_end_date ),
		$subscription,
		$new_ship_date->format( 'Y-m-d H:i:s' ),
		$old_ship_date->format( 'Y-m-d H:i:s' )
	);
	$new_end_date                = ( ! empty( $old_end_date ) ) ? clone $old_end_date : null;

	if ( $adjust_end_date ) {
		// Normalize time for consistency.
		$old_ship_date->setTime( $time['hour'], $time['minute'], $time['second'] );

		$ship_diff    = $old_ship_date->diff( $new_ship_date );
		$new_end_date = DateTime::createFromFormat( 'Y-m-d H:i:s', $subscription->get_date( 'end', 'site' ), $time_zone );

		if ( $new_end_date ) {
			$old_end_date = $new_end_date->format( 'Y-m-d H:i:s' );

			$new_end_date->add( $ship_diff );
			$new_end_date->setTime( $time['hour'], $time['minute'], $time['second'] );
		}

		/**
		 * Filter for end date when it is changed.
		 *
		 * @param DateTime $new_end_date
		 * @param \WC_Subscription $subscription
		 * @param int $user_id
		 *
		 * @since 1.4.24
		 */
		$new_end_date = apply_filters( 'sptb_new_end_date', $new_end_date, $subscription, $user_id );

		if ( $new_end_date && $new_ship_date->getTimestamp() > $new_end_date->getTimestamp() ) {
			wc_add_notice( apply_filters( 'jgtb_date_renewal_error_message', __( 'Next shipment date couldn\'t be updated. Next payment date is inferior to the subscription end date.', 'jg-toolbox' ), $subscription ), 'error' );
			return 0;
		}

		$dates['end'] = $new_end_date ? $new_end_date->format( 'Y-m-d H:i:s' ) : $new_end_date;

		if ( $new_end_date ) {
			$note .= sprintf( '<br>&nbsp;<br>The end date was also changed<br>from %s<br>to %s.', $old_end_date, $new_end_date->format( 'Y-m-d H:i:s' ) );
		}
	}

	// Final validation to prevent future date errors.
	if ( $new_end_date && $new_ship_date->getTimestamp() > $new_end_date->getTimestamp() ) {
		wc_add_notice( apply_filters( 'jgtb_date_renewal_error_message', __( 'Next shipment date couldn\'t be updated. Next payment date is inferior to the subscription end date.', 'jg-toolbox' ), $subscription ), 'error' );
		return 0;
	}

	$note .= '</p>';

	$subscription->update_dates( $dates, 'site' );
	$subscription->add_order_note( $note, false, true );

	do_action( 'jgtb_after_change_next_ship_date', $subscription, $user_id, $old_ship_date, $new_ship_date, $old_end_date, $adjust_end_date );

	wc_add_notice( apply_filters( 'jgtb_date_renewal_successful_message', __( 'Next shipment date has been successfully updated', 'jg-toolbox' ), $subscription ), 'success' );

	// Update subscription last modified date.
	Utilities\update_post_modified_date( $subscription );

	$subscription->save();

	return 1;
}
