<?php
namespace Javorszky\Toolbox\Utilities;

use DateTime;
use DateTimeZone;
use WC_Subscriptions;
use WC_Subscription;

add_action( 'wp_enqueue_scripts', __NAMESPACE__ . '\\add_js' );
add_filter( 'jgtb_change_placeholder_dates', __NAMESPACE__ . '\\change_date_format' );
add_action( 'woocommerce_subscription_payment_complete', __NAMESPACE__ . '\\maybe_remove_skip', 10 );

/**
 * General method that handles adding markup. Separate because this should also be used to enqueue any javascript
 * files if we need them (for example date pickers).
 */
function add_js() {
	wp_register_script( 'jgtb', JGTB_URL . 'assets/js/jgtb.js', array( 'jquery', 'jquery-ui-datepicker' ), JGTB_VERSION, array( 'in_footer' => true ) );
	wp_enqueue_script( 'jgtb' );
	wp_localize_script(
		'jgtb',
		'wcs_toolbox',
		array(
			'text' => array(
				'cancel'               => __( 'Cancel', 'jg-toolbox' ),
				'yes'                  => __( 'Yes', 'jg-toolbox' ),
				'skip_next'            => __( 'Are you sure you want to skip next payment?', 'jg-toolbox' ),
				'ship_now_keep_date'   => __( 'Are you sure you want to ship now and keep schedule?', 'jg-toolbox' ),
				'ship_now_recalculate' => __( 'Are you sure you want to ship now and reschedule from today?', 'jg-toolbox' ),
			),
			'ship_date_format' => ship_date_format( 'js' ),
		)
	);
	$css = 'jquery-ui.min.css';
	if ( defined( 'WP_DEBUG' ) && WP_DEBUG ) {
		$css = 'jquery-ui.css';
	}

	wp_register_style( 'jquery-ui-datepicker', JGTB_URL . 'assets/css/' . $css, array(), JGTB_VERSION );

	wp_enqueue_style( 'jquery-ui-datepicker' );

	if ( is_account_page() ) {
		wp_register_style( 'toolbox', JGTB_URL . 'assets/css/toolbox.css', array(), JGTB_VERSION );
		wp_enqueue_style( 'toolbox' );
	}
}

/**
 * Strips the query args from the current URL, so clicking on buttons won't accidentally trigger other actions than
 * we wanted.
 *
 * @return string                           The current URL minus a few query args
 */
function strip_custom_query_args() {
	return remove_query_arg( array( 'skip_next_shipping', 'ship_now_keep_date', 'ship_now_adjust_date', 'new_frequency' ) );
}

function generate_nonce_on_items( $subscription ) {
	$string = '';
	foreach ( $subscription->get_items() as $item ) {
		$string .= $item['product_id'] . '-' . $item['qty'] . '-' . $item['variation_id'];
	}
	return $string;
}

/**
 * This will calculate the next payment date from today. Today, because we're calling "ship_now" previously, which
 * sets the "last_payment" as "now".
 *
 * @param \WC_Subscription $subscription   Subscription object we're changing the date on
 * @param string $old_next_date            Next date before we changed it. It's in UTC, and format 'Y-m-d H:i:s'
 * @return mixed                           Next calculated date.
 * @throws \Exception
 */
function adjust_date( $subscription, $old_next_date ) {
	if ( is_ship_reschedule_from_next_payment_available() ) {
		$next_payment_timestamp = wcs_add_time( $subscription->get_billing_interval(), $subscription->get_billing_period(), wcs_date_to_time( $old_next_date ), 'offset_site_time' );

		// Make sure the next payment is more than 2 hours in the future, this ensures changes to the site's timezone because of daylight savings will never cause a 2nd renewal payment to be processed on the same day
		$i = 1;
		while ( $next_payment_timestamp < ( time() + 2 * HOUR_IN_SECONDS ) && $i < 3000 ) {
			$next_payment_timestamp = wcs_add_time( $subscription->get_billing_interval(), $subscription->get_billing_period(), $next_payment_timestamp, 'offset_site_time' );

			$i++;
		}

		if ( $next_payment_timestamp > 0 ) {
			$next_payment_date = gmdate( 'Y-m-d H:i:s', $next_payment_timestamp );
		} else {
			$next_payment_date = 0;
		}
	} else {
		$next_payment_date = $subscription->calculate_date( 'next_payment' );
	}

	$cancelled_date = $subscription->get_date( 'cancelled' );

	if ( $cancelled_date && $next_payment_date > $cancelled_date ) {
		$next_payment_date = $cancelled_date;
	}

	$dates = array(
		'next_payment' => $next_payment_date,
	);

	if ( apply_filters( 'jgtb_adjust_end_date_with_next_date', false, $subscription, $next_payment_date, $old_next_date ) ) {
		// let's calculate the difference between the two
		$utc = new DateTimeZone( 'UTC' );

		$old_next_date_object = DateTime::createFromFormat( 'Y-m-d H:i:s', $old_next_date, $utc );
		$new_next_date_object = DateTime::createFromFormat( 'Y-m-d H:i:s', $next_payment_date, $utc );
		$now                  = new DateTime( 'now', $utc );
		$old_end_date         = $subscription->get_date( 'end' );
		$end_date_object      = DateTime::createFromFormat( 'Y-m-d H:i:s', $old_end_date, $utc );

		$diff = $old_next_date_object->diff( $now );

		$end_date_object->add( $diff );

		$dates['end']          = $end_date_object->format( 'Y-m-d H:i:s' );
		$dates['next_payment'] = ( $new_next_date_object >= $end_date_object ) ? 0 : $new_next_date_object->format( 'Y-m-d H:i:s' );
	}

	$subscription->update_dates( $dates );
	$subscription->save();
	return $next_payment_date;
}

/**
 * This will create a renewal order, but otherwise leave the next date intact. Used in both "ship now" requests.
 *
 * @param  \WC_Subscription  $subscription   subscription object we're generating a renewal order for
 */
function ship_now( $subscription ) {
	do_action( 'woocommerce_scheduled_subscription_payment', $subscription->get_id() );
}

/**
 * Extracts address details in posted data. Plucks from $_REQUEST
 *
 * @param  array        $request    request data (usually $_POST)
 * @param  string       $type       type of address details to pluck
 * @return array                    plucked address details
 */
function sieve_address_details( $request, $type = 'billing' ) {
	$type    = ( 'billing' === $type ) ? 'billing' : 'shipping';
	$address = array();
	foreach ( $request as $key => $value ) {
		if ( 0 === strpos( $key, $type ) ) {
			$new_key             = str_replace( $type . '_', '', $key );
			$address[ $new_key ] = $value;
		}
	}
	return $address;
}

/**
 * Returns an array containing all values that have changed in new.
 * @param  array        $old        Existing values to check against
 * @param  array        $new        New values to see whether anything changed
 * @return array                    Containing everything from New that changed with regards to Old
 */
function get_difference( $old, $new ) {
	$diff = array();
	foreach ( $new as $key => $value ) {
		if ( ! array_key_exists( $key, $old ) || $old[ $key ] !== $value ) {
			$diff[ $key ] = $value;
		}
	}
	return $diff;
}

/**
 * Makes sure that the date is a date, that it's not in the past, and that it's at least 24 hours in the future due
 * to wanting to avoid rescheduling bugs.
 *
 * @param  string           $new_date       ideally a mysql date format
 * @return boolean                          true if all is well and we can change the date on the subscription
 */
function sanitize_new_date( $new_date ) {
	$result    = true;
	$timestamp = strtotime( $new_date );

	// because incoming timestamp is going to be only YYYY-MM-DD, it would default to 0:00:00, and we need 3 am for
	// consistency
	$timestamp += 3 * HOUR_IN_SECONDS;

	if ( ! $timestamp ) {
		wc_add_notice( _x( 'New payment date is in an invalid format. Please check again.', 'Error notice while sanitizing date format', 'jg-toolbox' ), 'error' );
		$result = false;
	} elseif ( $timestamp < time() ) {
		wc_add_notice( _x( 'New payment date is in the past.', 'Error notice when next payment is in the past', 'jg-toolbox' ), 'error' );
		$result = false;
	} elseif ( $timestamp < time() + DAY_IN_SECONDS ) {
		wc_add_notice( _x( 'New payment date is too close to today, it might lead to renewal errors. Please use one of the "ship now" options.', 'Error notice when validating new next payment date', 'jg-toolbox' ), 'error' );
		$result = false;
	}

	return $result;
}

function is_feature_available( $feature ) {
	if ( 'yes' === get_option( JGTB_OPTION_PREFIX . $feature ) ) {
		return true;
	}
	return false;
}

function is_skip_next_available() {
	return is_feature_available( 'skip_next_available' );
}

function is_ship_keep_available() {
	return is_feature_available( 'ship_now_keep_available' );
}

function is_ship_reschedule_from_next_payment_available() {
	return is_feature_available( 'reschedule_from_payment_date' );
}

function is_ship_reschedule_available() {
	return is_feature_available( 'ship_now_reschedule_available' );
}

function is_edit_details_available() {
	return is_feature_available( 'edit_sub_details_available' );
}

function is_change_next_payment_available() {
	return is_feature_available( 'change_next_payment_available' );
}

function is_add_to_subscription_available() {
	return is_feature_available( 'add_to_subscription' );
}

function is_bulk_edit_available() {
	return is_feature_available( 'bulk_edit_subscriptions' );
}

/**
 * Implements functionality to replace placeholder text in button / description texts.
 *
 * @param  string           $button_text   text used in buttons
 * @param  \WC_Subscription  $subscription  subscription the dates relate to
 * @return string                          button text but with the placeholder replaced
 */
function replace_key_dates( $button_text, $subscription ) {
	add_filter( 'wcs_calculate_next_payment_from_last_payment', '__return_false' );
	$next_payment_date = $subscription->calculate_date( 'next_payment' );
	remove_filter( 'wcs_calculate_next_payment_from_last_payment', '__return_false' );

	$search = array(
		'[date_created]',
		'[start_date]',
		'[next_date]',
		'[next_date_from_next]',
		'[next_date_from_last]',
		'[next_date_from_today]',
	);

	$replace = apply_filters(
		'jgtb_change_placeholder_dates',
		array(
			$subscription->get_date( 'date_created' ),
			$subscription->get_date( 'date_created' ),
			$subscription->get_date( 'next_payment' ),
			$next_payment_date,
			$subscription->calculate_date( 'next_payment' ),
			calculate_next_payment_date_from_today( $subscription ),
		)
	);

	return str_replace( $search, $replace, $button_text );
}

/**
 * Example function to turn standard MySQL datetime formats into some other format. In this case, the default
 * datetime format used by WooCommerce.
 *
 * @param  array    $dates   This should be an array of dates in MySQL format
 * @return array             array of same dates in a different format
 */
function change_date_format( $dates ) {
	return array_map(
		function( $date ) {
			$ts = strtotime( $date );

			return gmdate( apply_filters( 'jgtb_change_placeholder_date_format', wc_date_format() ), $ts );
		},
		$dates
	);
}

/**
 * This is a copy of part of the method from Subscriptions because Subs doesn't have a generic "calculate next payment
 * date from" method.
 *
 * @param  \WC_Subscription $subscription  Subscription we're getting the next payment date for from today
 * @return string                         datetime in MySQL format in GMT timezone
 */
function calculate_next_payment_date_from_today( $subscription ) {
	$end_time               = $subscription->get_time( 'end' );
	$from_timestamp         = time();
	$next_payment_timestamp = wcs_add_time( $subscription->get_billing_interval(), $subscription->get_billing_period(), $from_timestamp );

	// Make sure the next payment is more than 2 hours in the future, this ensures changes to the site's timezone because of daylight savings will never cause a 2nd renewal payment to be processed on the same day
	$i = 1;
	while ( $next_payment_timestamp < ( time() + 2 * HOUR_IN_SECONDS ) && $i < 3000 ) {
		$next_payment_timestamp = wcs_add_time( $subscription->get_billing_interval(), $subscription->get_billing_period(), $next_payment_timestamp );
		$i++;
	}

	// If the subscription has an end date and the next billing period comes after that, return 0
	if ( 0 !== $end_time && ( $next_payment_timestamp + 23 * HOUR_IN_SECONDS ) > $end_time ) {
		$next_payment_timestamp = 0;
	}

	$next_payment_date = 0;

	if ( $next_payment_timestamp > 0 ) {
		$next_payment_date = gmdate( 'Y-m-d H:i:s', $next_payment_timestamp );
	}

	return $next_payment_date;
}

/**
 * Gets button text / and default
 *
 * @param string $button the button option id without the prefix
 * @param string $default default string in case the option is not set or the string is empty
 * @return string
 */
function get_button_text( $button, $default = '' ) {

	// use get_option to return default value
	$button_text = get_option( JGTB_OPTION_PREFIX . $button, $default );

	// in case the button is saved but is empty, use the default anyways
	if ( '' === $button_text ) {
		$button_text = $default;
	}

	return $button_text;
}

/**
 * Updates post modified to now.
 *
 * @param $post_id
 */
function update_post_modified_date( $post_id ) {

	if ( is_a( $post_id, \WC_Subscription::class ) ) {
		$post_id = $post_id->get_id();
	}

	if ( absint( $post_id ) < 1 ) {
		return;
	}

	$subscription = wcs_get_subscription( $post_id );

	if ( ! $subscription ) {
		return;
	}

	$subscription->set_date_modified( current_time( 'mysql' ) );
}

/**
 * Create a string that represents an existing subscription. This is used in the add-to-existing-subscription template.
 * @param $subscription
 * @return string
 */
function generate_option_text( $subscription ) {
	// translators: $1: ID of subscription, $2: recurrence, eg "2 months", $3: next payment date
	$base_text = sprintf( esc_html__( '#%1$d - every %2$s - next: %3$s', 'jg-toolbox' ), $subscription->get_id(), wcs_get_subscription_period_strings( $subscription->get_billing_interval(), $subscription->get_billing_period() ), date_i18n( wc_date_format(), $subscription->get_time( 'next_payment' ) ) );

	return $base_text;
}

/**
 * Convert date to format for display in emails.
 * Format: 'F j, Y'
 *
 * @param string $date in format 'Y-m-d H:i:s'
 *
 * @return string
 */
function email_display_date( $date ) {
	$tz       = new DateTimeZone( wc_timezone_string() );
	$new_date = DateTime::createFromFormat( 'Y-m-d H:i:s', $date, $tz ); // phpcs:ignore CSRF ok

	$new_date = apply_filters( 'jgtb_email_display_date', $new_date );
	return $new_date->format( 'F j, Y' );
}

/**
 * After subscription renewal clear the skipped meta so the customer can skip again.
 *
 * @param WC_Subscription $subscription
 * @return void
 */
function maybe_remove_skip( $subscription ) {
	$subscription->delete_meta_data( 'wcs_toolbox_skipped' );
	$subscription->save_meta_data();
}

/**
 * Get the date format for the ship date.
 * 
 * jQuery date picker formats: https://api.jqueryui.com/datepicker/#utility-formatDate
 *
 * @param  string $lang
 * @return string
 */
function ship_date_format( $lang = 'php' ) {
	$formats = array(
		'php' => 'Y-m-d',
		'js'  => 'yy-mm-dd', // datepicker format
	);
	
	if ( in_array( $lang, array_keys( $formats ) ) ) {
		$format = $formats[ $lang ];
	} else {
		$format = $formats['php'];
	}
	
	return apply_filters( 'jgtb_ship_date_format', $format, $lang );
}

/**
 * Edit address page.
 *
 * @param object $subscription
 * @param string $load_address Type of address; 'billing' or 'shipping'.
 */
function edit_address( $subscription, $load_address = 'billing' ) {
	$load_address   = sanitize_key( $load_address );
	$country_getter = "get_{$load_address}_country";
	$country        = is_callable( array( $subscription, $country_getter ) ) ? $subscription->$country_getter() : false;
	
	if ( ! $country ) {
		$country = get_user_meta( get_current_user_id(), $load_address . '_country', true );
	}

	if ( ! $country ) {
		$country = \WC()->countries->get_base_country();
	}

	if ( 'billing' === $load_address ) {
		$allowed_countries = \WC()->countries->get_allowed_countries();

		if ( ! array_key_exists( $country, $allowed_countries ) ) {
			$country = current( array_keys( $allowed_countries ) );
		}
	}

	if ( 'shipping' === $load_address ) {
		$allowed_countries = \WC()->countries->get_shipping_countries();

		if ( ! array_key_exists( $country, $allowed_countries ) ) {
			$country = current( array_keys( $allowed_countries ) );
		}
	}

	$address              = \WC()->countries->get_address_fields( $country, $load_address . '_' );
	$subscription_address = $subscription->get_address( $load_address );

	// Enqueue scripts.
	wp_enqueue_script( 'wc-country-select' );
	wp_enqueue_script( 'wc-address-i18n' );

	// Prepare values.
	foreach ( $address as $key => $field ) {
		$subscription_key = str_replace( "{$load_address}_", '', $key );
		
		if ( isset( $subscription_address[ $subscription_key ] ) ) {
			$address[ $key ]['value'] = apply_filters( 'woocommerce_my_account_edit_address_field_value', $subscription_address[ $subscription_key ], $key, $load_address );
		}
	}
	
	wc_get_template(
		'myaccount/edit-subscription-address.php',
		array(
			'load_address' => $load_address,
			'address'      => apply_filters( 'woocommerce_address_to_edit', $address, $load_address ),
		),
		'',
		JGTB_PATH . 'templates/'
	);
}

/**
 * Get the shipping zones.
 *
 * @return array
 */
function get_shipping_zones(): array {
	// Shipping methods extracted from shipping zones and default zone processing.
	$zones_with_default    = \WC_Shipping_Zones::get_zones();
	$zones_with_default[0] = \WC_Shipping_Zones::get_zone( 0 ); // add the default zone at index 0
	
	return $zones_with_default;
}

/**
 * Helper function to generate shipping method title with legacy tag if applicable.
 *
 * @param \WC_Shipping_Method $shipping_method
 * 
 * @return string
 */
function get_shipping_method_title( \WC_Shipping_Method $shipping_method ): string {
	$title = '';

	if ( is_callable( array( $shipping_method, 'get_title' ) ) ) {
		$title = $shipping_method->get_title();
	}
	
	if ( empty( $title ) && is_callable( array( $shipping_method, 'get_method_title' ) ) ) {
		$title = $shipping_method->get_method_title();
	}

	if ( 'local_pickup' === $shipping_method->id && checkout_is_block() ) {
		$title .= ' (' . __( 'legacy', 'jg-toolbox' ) . ')';
	}
	
	// 'PickupLocation' object as 'cost' property protected
	if ( 'pickup_location' === $shipping_method->id ) {
		$pickup_locations_settings = class_exists( '\\Automattic\\WooCommerce\\StoreApi\\Utilities\\LocalPickupUtils' ) ? \Automattic\WooCommerce\StoreApi\Utilities\LocalPickupUtils::get_local_pickup_settings() : array();
		$cost                      = isset( $pickup_locations_settings['cost'] ) ? $pickup_locations_settings['cost'] : 0;
	} else {
		$cost                      = property_exists( $shipping_method, 'cost' ) && ! empty( $shipping_method->cost ) ? $shipping_method->cost : 0;
	}
	
	if ( ! empty( $cost ) ) {
		$cost = floatval( trim( $cost ) );
		
		if ( is_numeric( $cost ) ) {
			$cost            = number_format( $cost, 2, wc_get_price_decimal_separator(), wc_get_price_thousand_separator() );
			$currency_symbol = get_woocommerce_currency_symbol();
			$title          .= ' - ' . $currency_symbol . $cost;
		}
	}

	return apply_filters( 'jgtb_get_shipping_method_title', $title, $shipping_method );
}

/**
 * Helper function to generate the shipping method label.
 *
 * @param string $suffix
 * @param string $title
 * @param string $id
 * 
 * @return string
 */
function generate_shipping_method_label( string $suffix, string $title, string $id = '' ): string {
	$label = sprintf( '%s - %s', $title, $suffix );

	if ( '' === $id ) {
		return $label;
	}

	return sprintf( '%s (#%s)', $label, $id );
}

/**
 * Gets all subscription shipping methods instance IDs.
 * In case the shipping method is a default one, it will return the shipping method ID.
 * 
 * @param \WC_Subscription $subscription
 * 
 * @return array
 */
function get_subscription_shipping_methods( \WC_Subscription $subscription ): array {
	$shipping_methods      = array();
	$zone_shipping_methods = get_subscription_available_shipping_methods( $subscription );

	// Map subscription shipping methods
	foreach ( $subscription->get_shipping_methods() as $shipping_item ) {
		$method_id    = is_callable( array( $shipping_item, 'get_method_id' ) ) ? $shipping_item->get_method_id() : '';
		$method_title = is_callable( array( $shipping_item, 'get_name' ) ) ? $shipping_item->get_name() : '';
		$found        = false;

		// Attempt to find the correct instance ID by matching method title and ID
		foreach ( $zone_shipping_methods as $key => $zone_method ) {
			$zone_method_id    = ! empty( $zone_method->id ) ? $zone_method->id : '';
			$zone_method_title = ! empty( $zone_method->title ) ? $zone_method->title : '';
			$zone_method_title = empty( $zone_method_title ) && is_callable( array( $zone_method, 'get_title' ) ) ? $zone_method->get_title() : $zone_method_title;
			
			if ( $zone_method_id === $method_id && $zone_method_title === $method_title ) {
				$shipping_methods[] = $key;
				$found              = true;
				break;
			}
		}

		if ( ! $found ) {
			$shipping_methods[] = $method_id;
		}
	}

	return apply_filters( 'jgtb_get_subscription_shipping_methods', $shipping_methods, $subscription );
}

/**
 * Gets the single subscription shipping method rate ID if only one exists.
 *
 * @param \WC_Subscription $subscription
 * @return string|null
 */
function get_single_subscription_shipping_rate_id( \WC_Subscription $subscription ): ?string {
	$subscription_shipping_methods = get_subscription_shipping_methods( $subscription );

	if ( count( $subscription_shipping_methods ) === 1 ) {
		return reset( $subscription_shipping_methods );
	}

	return null;
}

/**
 * Gets all available pickup locations for the checkout block.
 * 
 * @return array
 */
function get_checkout_block_pickup_locations(): array {
	$pickup_locations = array();
	
	if ( ! checkout_is_block() ) {
		return $pickup_locations;
	}
	
	$pickup_locations_settings = class_exists( '\\Automattic\\WooCommerce\\StoreApi\\Utilities\\LocalPickupUtils' ) ? \Automattic\WooCommerce\StoreApi\Utilities\LocalPickupUtils::get_local_pickup_settings() : array();
	$pickup_locations_enabled  = isset( $pickup_locations_settings['enabled'] ) ? $pickup_locations_settings['enabled'] : false;
	$pickup_locations_title	   = isset( $pickup_locations_settings['title'] ) ? $pickup_locations_settings['title'] : __( 'Pickup Locations', 'jg-toolbox' );

	if ( $pickup_locations_enabled && class_exists( '\\Automattic\\WooCommerce\\Blocks\\Shipping\\PickupLocation' ) ) {
		$_pickup_locations = get_option( 'pickup_location_pickup_locations', array() );
		
		foreach ( $_pickup_locations as $index => $location ) {
			if ( wc_string_to_bool( $location['enabled'] ) ) {
				$pickup_location              = new \Automattic\WooCommerce\Blocks\Shipping\PickupLocation();
				$pickup_location->id          = 'pickup_location';
				$pickup_location->instance_id = $index;
				$pickup_location->title       = sprintf( '%s (%s)', $pickup_locations_title, $location['name'] );
				
				$pickup_locations[ 'pickup_location:' . $index ] = $pickup_location;
			}
		}
	}

	return apply_filters( 'jgtb_get_checkout_block_pickup_locations', $pickup_locations, $pickup_locations_settings );
}

/**
 * Gets all available shipping methods for a subscription.
 * 
 * @param \WC_Subscription $subscription
 * 
 * @return array
 */
function get_subscription_available_shipping_methods( \WC_Subscription $subscription ): array {
	$zone_shipping_methods = array();

	// Get the shipping zone for the subscription's shipping address
	$shipping_address = array(
		'country'   => $subscription->get_shipping_country(),
		'state'     => $subscription->get_shipping_state(),
		'postcode'  => $subscription->get_shipping_postcode(),
		'city'      => $subscription->get_shipping_city(),
		'address'   => $subscription->get_shipping_address_1(),
		'address_2' => $subscription->get_shipping_address_2(),
	);

	$shipping_zone = \WC_Shipping_Zones::get_zone_matching_package( array( 'destination' => $shipping_address ) );

	foreach ( $shipping_zone->get_shipping_methods() as $shipping_method ) {
		if ( ! isset( $shipping_method->id ) || ! isset( $shipping_method->instance_id ) || ! is_callable( array( $shipping_method, 'is_enabled' ) ) || ! $shipping_method->is_enabled() ) {
			continue;
		}

		$zone_shipping_methods[ $shipping_method->id . ':' . $shipping_method->instance_id ] = $shipping_method;
	}
	
	$zone_shipping_methods = array_merge( $zone_shipping_methods, get_checkout_block_pickup_locations() );
	
	return apply_filters( 'jgtb_get_subscription_available_shipping_methods', $zone_shipping_methods, $subscription );
}

/**
 * Get shipping methods as options.
 * 
 * @param \WC_Subscription $subscription
 * 
 * @return array
 */
function get_available_shipping_methods_as_options( \WC_Subscription $subscription ): array {
	$shipping_options           = array();
	$available_shipping_methods = get_subscription_available_shipping_methods( $subscription );
	
	foreach ( $available_shipping_methods as $key => $shipping_method ) {
		if ( ! isset( $shipping_method->id ) || ! isset( $shipping_method->instance_id ) || ! isset( $shipping_method->title ) ) {
			continue;
		}
		
		if ( $shipping_method instanceof \WC_Shipping_Method && is_callable( array( $shipping_method, 'get_instance_id' ) ) ) {
			$instance_id              = absint( $shipping_method->get_instance_id() );
			$zone                     = \WC_Shipping_Zones::get_zone_by( 'instance_id', absint( $instance_id ) );
			$zone_id                  = $zone && is_callable( array( $zone, 'get_id' ) ) ? $zone->get_id() : 0;
			$zone_name                = $zone && is_callable( array( $zone, 'get_zone_name' ) ) ? $zone->get_zone_name() : '';
			$shipping_method_title    = get_shipping_method_title( $shipping_method );
			$suffix                   = ( $zone_id === 0 && empty( $zone_name ) ) ? __( 'Other locations', 'jg-toolbox' ) : $zone_name;
			$shipping_options[ $key ] = generate_shipping_method_label( $suffix, $shipping_method_title, $instance_id );
		}
	}

	return apply_filters( 'jgtb_get_shipping_methods_as_options', $shipping_options, $available_shipping_methods );
}

/**
 * Determine if the checkout is a block.
 *
 * @return bool
 */
function checkout_is_block(): bool {
	$checkout_page_id = wc_get_page_id( 'checkout' );
	$is_block         = $checkout_page_id && function_exists( 'has_block' ) && has_block( 'woocommerce/checkout', $checkout_page_id );

	if ( ! $is_block ) {
		$is_block = class_exists( 'WC_Blocks_Utils' ) && count( \WC_Blocks_Utils::get_blocks_from_page( 'woocommerce/checkout', 'checkout' ) ) > 0;
	}

	return $is_block;
}

/**
 * Maybe recalculate and apply shipping for a subscription with a fallback to the original shipping method if needed.
 *
 * @param \WC_Subscription $subscription
 * @return void
 */
function maybe_recalculate_and_apply_shipping_for_subscription( \WC_Subscription $subscription ): void {
	if ( ! apply_filters( 'jgtb_allow_shipping_calc_on_product_update', wc_string_to_bool( get_option( JGTB_OPTION_PREFIX . 'shipping_calc_on_product_update', 'no' ) ), $subscription ) ) {
		return;
	}
	
	$original_shipping_items = array();

	// Store existing shipping items in case we need to revert
	foreach ( $subscription->get_items( 'shipping' ) as $item_id => $shipping_item ) {
		$original_shipping_items[] = $shipping_item;
		$subscription->remove_item( $item_id );
	}

	// Recalculate subscription totals to account for any changes (including coupons and discounts)
	$subscription->calculate_totals();
	
	$original_method_id = null;

	// Try to find the original shipping method
	if ( ! empty( $original_shipping_items ) ) {
		$original_method_id = $original_shipping_items[0]->get_method_id();
	}
	
	$original_method_id = is_numeric( $original_method_id ) ? absint( $original_method_id ) : null;

	// Try applying the original, forced, or fallback shipping method
	if ( ! apply_best_shipping_rate_to_subscription( $subscription, $original_method_id ) ) {
		// Reapply the original shipping items
		foreach ( $original_shipping_items as $original_item ) {
			$subscription->add_item( $original_item );
		}
	}

	// Recalculate totals again after applying shipping and save the subscription
	$subscription->calculate_totals();
	$subscription->save();
}

/**
 * Apply the best shipping rate to the subscription. It will try to apply the original method first,
 * then a forced method if available, and validate free shipping if needed.
 *
 * @param \WC_Subscription $subscription
 * @param string|null $original_method_id
 * 
 * @return bool True if a shipping method was applied, false otherwise.
 */
function apply_best_shipping_rate_to_subscription( \WC_Subscription $subscription, ?string $original_method_id ): bool {
	$available_shipping_methods = get_subscription_available_shipping_methods( $subscription );
	$forced_method_id           = apply_filters( 'jgtb_force_shipping_method', false, $subscription );

	// Check if the original method is free shipping
	if ( $original_method_id && check_shipping_method_exists( $available_shipping_methods, $original_method_id, 'free_shipping' ) ) {
		$free_shipping_method = get_shipping_method_by_id( $available_shipping_methods, $original_method_id );
		
		if ( $free_shipping_method && validate_free_shipping_method( $free_shipping_method, $subscription ) ) {
			add_new_shipping_item_to_subscription( $subscription, $free_shipping_method );
			return true; // Free shipping successfully applied
		}
	}

	// Check if the forced method is free shipping
	if ( $forced_method_id && check_shipping_method_exists( $available_shipping_methods, $forced_method_id, 'free_shipping' ) ) {
		$free_shipping_method = get_shipping_method_by_id( $available_shipping_methods, $forced_method_id );
		
		if ( $free_shipping_method && validate_free_shipping_method( $free_shipping_method, $subscription ) ) {
			add_new_shipping_item_to_subscription( $subscription, $free_shipping_method );
			return true; // Free shipping successfully applied
		}
	}
	
	// Remove any free_shipping methods from available methods array before fallback processing
	foreach ( $available_shipping_methods as $method_id => $method ) {
		if ( false !== strpos( $method->id, 'free_shipping' ) ) {
			unset( $available_shipping_methods[ $method_id ] );
		}
	}

	// Apply the original shipping method if available
	if ( $original_method_id && check_shipping_method_exists( $available_shipping_methods, $original_method_id ) ) {
		$shipping_method = get_shipping_method_by_id( $available_shipping_methods, $original_method_id );
		
		if ( $shipping_method ) {
			add_new_shipping_item_to_subscription( $subscription, $shipping_method );
			return true;
		}
	}

	// Apply the forced shipping method if available
	if ( $forced_method_id && check_shipping_method_exists( $available_shipping_methods, $forced_method_id ) ) {
		$shipping_method = get_shipping_method_by_id( $available_shipping_methods, $forced_method_id );
		
		if ( $shipping_method ) {
			add_new_shipping_item_to_subscription( $subscription, $shipping_method );
			return true;
		}
	}

	// Apply the first available rate as fallback
	$first_available_method = reset( $available_shipping_methods );
	if ( $first_available_method ) {
		add_new_shipping_item_to_subscription( $subscription, $first_available_method );
		return true;
	}

	// No shipping method was applied
	return false;
}

/**
 * Check if the given shipping method exists and optionally checks if it's a free shipping method.
 *
 * @param array $available_shipping_methods Array of available shipping methods.
 * @param string $method_id Method ID.
 * @param string|null $type Optional. Shipping method type to check for (e.g., 'free_shipping').
 * 
 * @return bool True if the method exists and matches the type, otherwise false.
 */
function check_shipping_method_exists( array $available_shipping_methods, string $method_id, ?string $type = null ): bool {
	foreach ( $available_shipping_methods as $id => $shipping_method ) {
		if ( false !== strpos( $id, $method_id ) && ( is_null( $type ) || $shipping_method->id === $type ) ) {
			return true;
		}
	}
	return false;
}

/**
 * Get a shipping method by its ID.
 *
 * @param array $available_shipping_methods Array of available shipping methods.
 * @param string $method_id Method ID.
 * 
 * @return \WC_Shipping_Method|null The shipping method if found, otherwise null.
 */
function get_shipping_method_by_id( array $available_shipping_methods, string $method_id ): ?\WC_Shipping_Method {
	foreach ( $available_shipping_methods as $id => $shipping_method ) {
		if ( false !== strpos( $id, $method_id ) ) {
			return $shipping_method;
		}
	}
	return null;
}

/**
 * Validate if free shipping method is applicable based on subscription total, coupons, etc.
 *
 * @param \WC_Shipping_Method $free_shipping_method The free shipping method.
 * @param \WC_Subscription $subscription The subscription object.
 * 
 * @return bool True if free shipping is applicable, false otherwise.
 */
function validate_free_shipping_method( \WC_Shipping_Method $free_shipping_method, \WC_Subscription $subscription ): bool {
	$min_amount         = (float) $free_shipping_method->get_option( 'min_amount', 0 );
	$requires           = $free_shipping_method->get_option( 'requires' );
	$ignore_discounts   = $free_shipping_method->get_option( 'ignore_discounts', 'no' );
	$has_met_min_amount = false;
	$has_coupon         = false;
	$total              = $subscription->get_subtotal();

	// Get the subscription contents cost
	if ( ! wc_string_to_bool( $ignore_discounts ) ) {
		$total -= $subscription->get_discount_total();
	}

	// Check if the total meets the minimum amount
	if ( $total >= $min_amount ) {
		$has_met_min_amount = true;
	}

	// Check if a valid free shipping coupon is applied
	foreach ( $subscription->get_coupon_codes() as $coupon_code ) {
		$coupon = new \WC_Coupon( $coupon_code );
		if ( $coupon->get_free_shipping() ) {
			$has_coupon = true;
			break;
		}
	}

	// Determine if free shipping is available based on the 'requires' option
	switch ( $requires ) {
		case 'min_amount':
			return $has_met_min_amount;
		case 'coupon':
			return $has_coupon;
		case 'both':
			return $has_met_min_amount && $has_coupon;
		case 'either':
			return $has_met_min_amount || $has_coupon;
		default:
			return true; // No specific requirements, so it's applicable
	}
}


/**
 * Helper function to add a new shipping method to a subscription.
 *
 * @param \WC_Subscription $subscription
 * @param \WC_Shipping_Method $shipping_method
 * 
 * @return void
 */
function add_new_shipping_item_to_subscription( \WC_Subscription $subscription, \WC_Shipping_Method $shipping_method ): void {
	$shipping_item = new \WC_Order_Item_Shipping();
	$method_title  = is_callable( array( $shipping_method, 'get_title' ) ) ? $shipping_method->get_title() : $shipping_method->title;
	$method_id     = is_callable( array( $shipping_method, 'get_method_id' ) ) ? $shipping_method->get_method_id() : $shipping_method->id;

	// Build the shipping package based on the subscription's details
	$shipping_address = array(
		'country'   => $subscription->get_shipping_country(),
		'state'     => $subscription->get_shipping_state(),
		'postcode'  => $subscription->get_shipping_postcode(),
		'city'      => $subscription->get_shipping_city(),
		'address'   => $subscription->get_shipping_address_1(),
		'address_2' => $subscription->get_shipping_address_2(),
	);

	$contents      = array();
	$contents_cost = 0;

	// Collect all products in the subscription that need shipping
	foreach ( $subscription->get_items() as $item_id => $item ) {
		$product = $item->get_product();

		if ( $product && $product->needs_shipping() ) {
			$contents[ $item_id ] = array(
				'data'       => $product,
				'quantity'   => $item->get_quantity(),
				'line_total' => $item->get_total(),
				'line_tax'   => $item->get_total_tax(),
			);
			
			$contents_cost += floatval( $item->get_total() );
		}
	}

	$package = array(
		'contents'        => $contents,
		'contents_cost'   => $contents_cost,
		'destination'     => $shipping_address,
		'applied_coupons' => $subscription->get_coupon_codes(),
	);

	// Calculate shipping for the given package
	\WC()->shipping()->reset_shipping(); // Reset any existing shipping calculations
	\WC()->shipping()->calculate_shipping( array( $package ) );
	$available_rates = \WC()->shipping()->get_packages()[0]['rates'];

	$shipping_item->set_props( array(
		'method_title' => $method_title,
		'method_id'    => $method_id,
		'total'        => 0,
		'taxes'        => array(),
		'order_id'     => $subscription->get_id(),
	) );

	// Check if the shipping method has rates after calculation
	foreach ( $available_rates as $rate_id => $rate ) {
		// Use strpos to match method IDs that contain instance IDs (e.g., flat_rate:1)
		if ( false !== strpos( $rate_id, $method_id ) ) {
			$shipping_item->set_props( array(
				'total' => wc_format_decimal( $rate->cost ),
				'taxes' => $rate->taxes,
			) );
			break;
		}
	}

	// Handle pickup location method if applicable
	if ( 'pickup_location' === $method_id ) {
		$local_pickup_settings = class_exists( '\\Automattic\\WooCommerce\\StoreApi\\Utilities\\LocalPickupUtils' ) ? \Automattic\WooCommerce\StoreApi\Utilities\LocalPickupUtils::get_local_pickup_settings() : array();
		$cost                  = isset( $local_pickup_settings['cost'] ) ? $local_pickup_settings['cost'] : '';
		
		if ( ! empty( $cost ) ) {
			$shipping_item->set_props( array(
				'total' => wc_format_decimal( $cost ),
				'taxes' => array(),
			) );
			$shipping_item->calculate_taxes();
		}
	}

	$shipping_item->save();
	$subscription->add_item( $shipping_item );
}




