<?php
namespace Javorszky\Toolbox\Utilities\Validate;

use Javorszky\Toolbox\Utilities as Utilities;

use WC_Subscription;

/**
 * Utility function that is used in all three validation scenarios for the common tasks of making sure there's a
 * subscription, and that the user actually owns the subscription.
 *
 * @param  int               $user_id        ID of the user performing the request
 * @param  \WC_Subscription  $subscription   the subscription object we're checking
 * @return bool                              whether the request is valid for the user / subscription
 */
function validate_subscription_ownership( int $user_id, \WC_Subscription $subscription ): bool {
	if ( ! wcs_is_subscription( $subscription ) ) {
		wc_add_notice( _x( 'That subscription does not exist. Please contact us if you need assistance.', 'Error notice while validating subscription ownership during all actions', 'jg-toolbox' ), 'error' );
		return false;

	} elseif ( ! user_can( $user_id, 'edit_shop_subscription_status', $subscription->get_id() ) ) {
		wc_add_notice( _x( 'This does not appear to be one of your subscriptions.', 'Error notice while validating subscription ownership during all actions.', 'jg-toolbox' ), 'error' );
		return false;
	}

	return true;
}

/**
 * Validates the request. Nonce needs to be correct so that we can't spam the same link but have to click the button
 * again (new nonce)
 *
 * User's ownership and subscription existing are already verified.
 *
 * @param  \WC_Subscription  $subscription  the subscription we're skipping next payment on
 * @param  int               $next_date     timestamp for when the previous next payment was. New next payment will be calculated from this
 * @param  string            $wpnonce       nonce generated based on subscription id and previous next payment timestamp
 * @return bool                             true if the request is right
 */
function validate_skip_next_request( \WC_Subscription $subscription, int $next_date, string $wpnonce ): bool {
	$action     = $subscription->get_id() . $next_date;
	$error_code = 'E01';
	$message    = sprintf(
		/* translators: %s is the error code */
		_x( 'Security error. Please contact us if you need assistance. %s.', 'Error notice when a skip next request fails nonce verification', 'jg-toolbox' ),
		$error_code
	);

	return validate_nonce( $wpnonce, $action, $error_code, $message );
}


/**
 * Validates the request part specific to the ship now keep date part. The nonce needs to be correct for completed
 * payment count and subscription ID to guard against spamming the link to charge the user multiple times.
 *
 * @param  \WC_Subscription  $subscription       the subscription we're skipping next payment on
 * @param  int               $completed_payments number of times the subscription's been paid for
 * @param  string            $wpnonce            nonce generated based on subscription id and previous next payment timestamp
 * @return bool                                  true if the request is right
 */
function validate_ship_now_keep_date( \WC_Subscription $subscription, int $completed_payments, string $wpnonce ): bool {
	$action     = $subscription->get_id() . '_completed_' . $completed_payments;
	$error_code = 'E02';
	$message    = sprintf(
		/* translators: %s is the error code */
		_x( 'Security error. Please contact us if you need assistance. %s.', 'Error notice when a ship now keep date fails nonce verification', 'jg-toolbox' ),
		$error_code
	);

	return validate_nonce( $wpnonce, $action, $error_code, $message );
}

/**
 * Validates the request part specific to the ship now adjust date part. The nonce needs to be correct for completed
 * payment count and subscription ID to guard against spamming the link to charge the user multiple times.
 *
 * @param  \WC_Subscription  $subscription       the subscription we're skipping next payment on
 * @param  int               $completed_payments number of times the subscription's been paid for
 * @param  string            $wpnonce            nonce generated based on subscription id and previous next payment timestamp
 * @return bool                                  true if the request is right
 */
function validate_ship_now_adjust_date( \WC_Subscription $subscription, int $completed_payments, string $wpnonce ): bool {
	$action     = $subscription->get_id() . '_completed_adjust_' . $completed_payments;
	$error_code = 'E03';
	$message    = sprintf(
		/* translators: %s is the error code */
		_x( 'Security error. Please contact us if you need assistance. %s.', 'Error notice when ship now adjust date fails nonce verification', 'jg-toolbox' ),
		$error_code
	);

	return validate_nonce( $wpnonce, $action, $error_code, $message );
}

/**
 * Make sure that our nonce is in order.
 *
 * @param  \WC_Subscription  $subscription  the subscription we're operating on
 * @param  string            $wpnonce       the nonce
 * @return bool                             true if the nonce is in order
 */
function validate_edit_details( \WC_Subscription $subscription, string $wpnonce ): bool {
	$action     = 'wcs_edit_details_of_' . $subscription->get_id();
	$error_code = 'E04';
	$message    = sprintf(
		/* translators: %s is the error code */
		_x( 'Security error. Please contact us if you need assistance. %s.', 'Error notice when edit details fails nonce verification', 'jg-toolbox' ),
		$error_code
	);

	return validate_nonce( $wpnonce, $action, $error_code, $message );
}

/**
 * Make sure that our nonce is in order.
 *
 * @param  \WC_Subscription  $subscription  the subscription we're operating on
 * @param  string            $wpnonce       the nonce
 * @return bool                             true if the nonce is in order
 */
function validate_change_next_ship_date_request( \WC_Subscription $subscription, string $wpnonce ): bool {
	$action     = 'change_next_ship_date_' . $subscription->get_id() . $subscription->get_date( 'next_payment', 'site' );
	$error_code = 'E05';
	$message    = sprintf(
		/* translators: %s is the error code */
		_x( 'Security error. Please contact us if you need assistance. %s.', 'Notice text when a validation error happens', 'jg-toolbox' ),
		$error_code
	);

	return validate_nonce( $wpnonce, $action, $error_code, $message );
}

/**
 * Makes sure that the same request cannot be replayed to add infinite products to the same subscription. It's
 * based on the items already in the subscription, so it will change after every "add to existing" request.
 *
 * @param  \WC_Subscription  $subscription  the subscription we're adding to
 * @param  string            $wpnonce       security nonce belonging to the subscription
 * @return bool                             whether the nonce is right for the subscription
 */
function validate_add_to_subscription_request( \WC_Subscription $subscription, string $wpnonce ): bool {
	$items_string = Utilities\generate_nonce_on_items( $subscription );
	$action       = 'add_to_subscription_' . $items_string;
	$error_code   = 'E06';
	$message      = sprintf(
		_x( 'Security error. Please contact us if you need assistance. %s.', 'Notice text when a validation error happens', 'jg-toolbox' ),
		$error_code
	);

	return validate_nonce( $wpnonce, $action, $error_code, $message );
}

/**
 * Validate the request to change the frequency.
 *
 * @param  \WC_Subscription  $subscription
 * @param  string            $wpnonce
 * @return bool
 */
function validate_change_frequency_request( \WC_Subscription $subscription, string $wpnonce ): bool {
	$action     = 'change_frequency_' . $subscription->get_id() . $subscription->get_billing_interval() . $subscription->get_billing_period();
	$error_code = 'E07';
	$message    = sprintf(
		_x( 'Security error. Please contact us if you need assistance. %s.', 'Notice text when a validation error happens', 'jg-toolbox' ),
		$error_code
	);

	return validate_nonce( $wpnonce, $action, $error_code, $message );
}

/**
 * Validate the request to change the shipping method.
 *
 * @param  \WC_Subscription  $subscription
 * @param  string            $wpnonce
 * @param  string            $old_shipping_method_rate_id
 * @return bool
 */
function validate_change_shipping_method_request( \WC_Subscription $subscription, string $wpnonce, string $old_shipping_method_rate_id ): bool {
    $action     = 'change_shipping_method_' . $subscription->get_id() . '_' . $old_shipping_method_rate_id;
    $error_code = 'E08';
    $message    = sprintf(
        _x( 'Security error. Please contact us if you need assistance. %s.', 'Notice text when a validation error happens', 'jg-toolbox' ),
        $error_code
    );

    return validate_nonce( $wpnonce, $action, $error_code, $message );
}

/**
 * Helper function to validate a nonce and log errors if it fails.
 *
 * @param string $wpnonce  The nonce to validate.
 * @param string $action   The action to compare against.
 * @param string $error_code The error code to log and display.
 * @param string $error_message The error message template.
 *
 * @return bool  True if the nonce is valid, false otherwise.
 */
function validate_nonce( string $wpnonce, string $action, string $error_code, string $error_message ): bool {
	$logger  = wc_get_logger();
	$context = array( 'source' => 'jgtb_validation' );

	if ( empty( $wpnonce ) ) {
		$logger->error( 'Nonce was empty: ' . $error_code, $context );
		wc_add_notice( $error_message, 'error' );
		return false;
	}

	if ( ! wp_verify_nonce( $wpnonce, $action ) ) {
		$logger->error(
			sprintf(
				'Verification failed for nonce %1$s and action %2$s: %3$s',
				$wpnonce,
				$action,
				$error_code
			),
			$context
		);
		wc_add_notice( $error_message, 'error' );
		return false;
	}

	return true;
}

/**
 * Validate the new next shipping date.
 *
 * @param  \DateTime       $new_next_ship_date
 * @param  \DateTime       $current_next_ship_date
 * @param  \DateTime|null  $end_date
 * @return \DateTime|null
 */
function validate_new_next_shipping_date( \DateTime $new_next_ship_date, \DateTime $current_next_ship_date, ?\DateTime $end_date = null ): ?\DateTime {
	if ( empty( $end_date ) ) {
		return $new_next_ship_date;
	}
	
	if ( $current_next_ship_date->format( 'Y-m-d' ) === $new_next_ship_date->format( 'Y-m-d' ) ) {
		return null;
	}
	
	$allow_date_after_end        = get_option( JGTB_OPTION_PREFIX . 'change_next_payment_available_after_end_date', 'no' );
	$allow_date_after_end_months = get_option( JGTB_OPTION_PREFIX . 'change_next_payment_available_after_end_date_months', 6 );
	$min                         = clone $current_next_ship_date;
	$max                         = clone $min;
	
	if ( $end_date && $allow_date_after_end && ! empty( $allow_date_after_end_months ) && is_numeric( $allow_date_after_end_months ) ) {
		$months = absint( $allow_date_after_end_months );
		$max    = clone $end_date;
		$max->modify( "+{$months} months" );
	}
	
	if ( $new_next_ship_date->getTimestamp() > $min->getTimestamp() && $new_next_ship_date->getTimestamp() <= $max->getTimestamp() ) {
		return $new_next_ship_date;
	} else {
		return null;
	}
}
